// vgmpack
// Mic, 2009

#include <cstdlib>
#include <iostream>
#include <vector>
#include <string>
#include <time.h>
#include "dictcodec.h"
#include "rlecodec.h"

using namespace std;


void pack(char *inFileName, char *outFileName, int mode)
{
    unsigned int i, inSize, vgmOffs, loopOffs, newLoopOffs, progress, extraDataOffs;
	FILE *inFile, *outFile;
	unsigned char c;
	unsigned char *vgmData;
	vector<unsigned char> outData;
	vector<unsigned char> extraDataBlock;
	unsigned short *dictionary;
	Codec *codec;
	bool endOfVgmData = false;

	fopen_s(&inFile, inFileName, "rb");
	if (inFile == NULL)
	{
		printf("Error: Failed to open %s\n", inFileName);
		exit(0);
	}
	fseek(inFile, 0, SEEK_END);
	inSize = ftell(inFile);
	fseek(inFile, 0, SEEK_SET);
    if ((vgmData = new unsigned char[inSize]) == NULL)
	{
		puts("Error: Failed to allocate memory");
	}
	if (fread(vgmData, 1, inSize, inFile) != inSize)
	{
		puts("Error: Failed to read VGM data");
		exit(0);
	}
	fclose(inFile);

	if ((vgmData[8] != 0x50) || (vgmData[9] != 0x01) ||
		(vgmData[10] != 0x00) || (vgmData[11] != 0x00))
	{
		puts("Error: Only VGM version 1.50 supported");
		exit(0);
	}
	vgmData[8] = 0x51;	// To identify the VGM as compressed

	vgmOffs = vgmData[0x34] +
		      (vgmData[0x35] << 8) +
			  (vgmData[0x36] << 16) +
			  (vgmData[0x37] << 24) + 0x34;
	extraDataOffs = vgmOffs;

	loopOffs = vgmData[0x1C] +
		       (vgmData[0x1D] << 8) +
			   (vgmData[0x1E] << 16) +
			   (vgmData[0x1F] << 24) + 0x1C;
	newLoopOffs = loopOffs;

	printf("Packing %s:", inFileName);

	for (i = 0; i < vgmOffs; i++)
	{
		outData.push_back(vgmData[i]);
	}

	if (mode == 1)
	{
		codec = new RleCodec(&outData);
	}
	else
	{
		codec = new DictCodec(&outData);
	}

	progress = 0;

	while (!endOfVgmData)
	{
		if (((vgmOffs * 10) / inSize) > progress)
		{
			progress++;
			printf("%c", 177);
		}

		c = vgmData[vgmOffs++];

		if (vgmOffs == loopOffs)
		{
			codec->flush();
			newLoopOffs = (unsigned int)outData.size();
		}

		codec->write(c);

		switch (c)
		{
			case 0x4F:
			case 0x50:
				codec->passThrough(vgmData[vgmOffs++]);
				break;

			case 0x51:
			case 0x52:
			case 0x53:
			case 0x54:
			case 0x61:
				codec->passThrough(vgmData[vgmOffs++]);
				codec->passThrough(vgmData[vgmOffs++]);
				break;

			case 0x66:
				endOfVgmData = true;
				break;

			case 0x67:
				if (vgmData[vgmOffs] == 0x66)
				{
					unsigned int dataSize = vgmData[vgmOffs+2];
					dataSize += vgmData[vgmOffs+3] << 8;
					dataSize += vgmData[vgmOffs+4] << 16;
					dataSize += vgmData[vgmOffs+5] << 24;
					for (i = 0; i < dataSize + 6; i++)
					{
						codec->passThrough(vgmData[vgmOffs++]);
					}
				}
				else
				{
					printf("Illegal command: 0x67 0x%02x\n", vgmData[vgmOffs]);
					delete [] vgmData;
					delete codec;
					exit(0);
				}
				break;

			case 0xE0:
				codec->passThrough(vgmData[vgmOffs++]);
				codec->passThrough(vgmData[vgmOffs++]);
				codec->passThrough(vgmData[vgmOffs++]);
				codec->passThrough(vgmData[vgmOffs++]);
				break;
		}
	}

	if (mode == 2)
	{
		extraDataBlock.push_back(0x67);
		extraDataBlock.push_back(0x66);
		extraDataBlock.push_back(0x01);
		extraDataBlock.push_back(DICT_SIZE & 0xFF);
		extraDataBlock.push_back(DICT_SIZE >> 8);
		extraDataBlock.push_back(0x00);
		extraDataBlock.push_back(0x00);
		dictionary = (unsigned short *)(codec->get(0));
		for (i = 0; i < DICT_SIZE; i++)
		{
			extraDataBlock.push_back((unsigned char)dictionary[i]);
		}
	}

	delete codec;

	// EOF offset
	i = (unsigned int)(outData.size() + extraDataBlock.size() - 4);
	outData[4] = i & 0xFF;
	outData[5] = (i >> 8) & 0xFF;
	outData[6] = (i >> 16) & 0xFF;
	outData[7] = (i >> 24) & 0xFF;

	// Read rest of data, if any (GD3)
	while (vgmOffs < inSize)
	{
		outData.push_back(vgmData[vgmOffs++]);
	}

	// GD3 offset
	i = outData[0x14] +
		(outData[0x15] << 8) +
		(outData[0x16] << 16) +
		(outData[0x17] << 24);
	if (i)
	{
		i -= (inSize - (outData.size() + extraDataBlock.size()));
		outData[0x14] = i & 0xFF;
		outData[0x15] = (i >> 8) & 0xFF;
		outData[0x16] = (i >> 16) & 0xFF;
		outData[0x17] = (i >> 24) & 0xFF;
	}

	// Loop offset
	if (loopOffs > 0x1C)
	{
		newLoopOffs += (unsigned int)extraDataBlock.size();
		newLoopOffs -= 0x1C;
		outData[0x1C] = newLoopOffs & 0xFF;
		outData[0x1D] = (newLoopOffs >> 8) & 0xFF;
		outData[0x1E] = (newLoopOffs >> 16) & 0xFF;
		outData[0x1F] = (newLoopOffs >> 24) & 0xFF;
	}

	while (progress < 10)
	{
		printf("%c", 177);
		progress++;
	}

	printf("\nInput size: %d bytes, output size: %d bytes\n", inSize, outData.size() + extraDataBlock.size());

	fopen_s(&outFile, outFileName, "wb");
	if (outFile == NULL)
	{
		delete [] vgmData;
		printf("Error: unable to open %s\n", outFileName);
	}
	else
	{
		for (i = 0; i < extraDataOffs; i++)
		{
			fputc(outData[i], outFile);
		}
		for (unsigned int j = 0; j < extraDataBlock.size(); j++)
		{
			fputc(extraDataBlock[j], outFile);
		}
		for (; i < outData.size(); i++)
		{
			fputc(outData[i], outFile);
		}
		fclose(outFile);
	}

	delete [] vgmData;
	outData.clear();
}


int main(int argc, char *argv[])
{
	int i, mode = 1;
	char *inFn = NULL, *outFn = NULL;

	puts("VGM Packer by Mic, 2009");

	for (i = 1; i < argc; i++)
	{
		if (argv[i][0] == '-')
		{
			if ((strcmp(argv[i], "-h")==0) ||
				(strcmp(argv[i], "-help")==0) ||
				(strcmp(argv[i], "-?")==0))
			{
				printf("Usage: vgmpack [options] <input> <output>\n");
				printf("Options:\n\t-1\tRun-length compression\n\t-2\tDictionary compression\n");
				return 0;
			}
			else if (strcmp(argv[i], "-1")==0)
			{
				mode = 1;
			}
			else if (strcmp(argv[i], "-2")==0)
			{
				mode = 2;
			}
		}
		else if (inFn == NULL)
		{
			inFn = argv[i];
		}
		else if (outFn == NULL)
		{
			outFn = argv[i];
		}
	}

	if ((inFn == NULL) || (outFn == NULL))
	{
		printf("Usage: vgmpack [options] <input> <output>\n");
		printf("Use -h for a description of the options\n");
        return 0;
    }

	pack(inFn, outFn, mode);

	return 0;
}


